from pathlib import Path
from typing import Literal

def get_project_imports(
    source_roots: list[str],
    file_path: str,
    ignore_type_checking_imports: bool,
) -> list[tuple[str, int]]: ...
def get_external_imports(
    source_roots: list[str],
    file_path: str,
    ignore_type_checking_imports: bool,
) -> list[tuple[str, int]]: ...
def get_normalized_imports(
    source_roots: list[str],
    file_path: str,
    ignore_type_checking_imports: bool,
) -> list[tuple[str, int]]: ...
def set_excluded_paths(
    project_root: str, exclude_paths: list[str], use_regex_matching: bool
) -> None: ...
def check_external_dependencies(
    project_root: str,
    source_roots: list[str],
    module_mappings: dict[str, list[str]],
    ignore_type_checking_imports: bool,
) -> tuple[dict[str, list[str]], dict[str, list[str]]]: ...
def create_dependency_report(
    project_root: str,
    source_roots: list[str],
    path: str,
    include_dependency_modules: list[str] | None,
    include_usage_modules: list[str] | None,
    skip_dependencies: bool,
    skip_usages: bool,
    ignore_type_checking_imports: bool,
) -> str: ...
def create_computation_cache_key(
    project_root: str,
    source_roots: list[str],
    action: str,
    py_interpreter_version: str,
    file_dependencies: list[str],
    env_dependencies: list[str],
    backend: str,
) -> str: ...
def check_computation_cache(
    project_root: str, cache_key: str
) -> tuple[list[tuple[int, str]], int] | None: ...
def update_computation_cache(
    project_root: str, cache_key: str, value: tuple[list[tuple[int, str]], int]
) -> None: ...
def parse_project_config(filepath: Path) -> ProjectConfig: ...
def parse_interface_members(source_roots: list[Path], path: str) -> list[str]: ...
def dump_project_config_to_toml(project_config: ProjectConfig) -> str: ...
def check(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str],
) -> CheckResult: ...
def sync_dependency_constraints(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str],
    prune: bool,
) -> ProjectConfig: ...
def sync_project(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str],
    add: bool = False,
) -> str: ...

class TachCircularDependencyError(ValueError):
    dependencies: list[str]

class TachVisibilityError(ValueError):
    visibility_errors: list[tuple[str, str, list[str]]]

class ErrorInfo:
    def is_dependency_error(self) -> bool: ...
    def to_pystring(self) -> str: ...
    def is_deprecated(self) -> bool: ...

class BoundaryError:
    file_path: Path
    line_number: int
    import_mod_path: str
    error_info: ErrorInfo

class CheckResult:
    errors: list[BoundaryError]
    deprecated_warnings: list[BoundaryError]
    warnings: list[str]

class DependencyConfig:
    path: str
    deprecated: bool

class ModuleConfig:
    path: str
    depends_on: list[DependencyConfig]
    visibility: list[str]
    strict: bool

    @staticmethod
    def new_root_config() -> ModuleConfig: ...
    def __new__(cls, path: str, strict: bool) -> ModuleConfig: ...
    def mod_path(self) -> str: ...

CacheBackend = Literal["disk"]

class CacheConfig:
    backend: CacheBackend
    file_dependencies: list[str]
    env_dependencies: list[str]

class ExternalDependencyConfig:
    exclude: list[str]

class UnusedDependencies:
    path: str
    dependencies: list[DependencyConfig]

class ProjectConfig:
    modules: list[ModuleConfig]
    cache: CacheConfig
    external: ExternalDependencyConfig
    exclude: list[str]
    source_roots: list[Path]
    exact: bool
    disable_logging: bool
    ignore_type_checking_imports: bool
    forbid_circular_dependencies: bool
    use_regex_matching: bool

    def __new__(cls) -> ProjectConfig: ...
    def module_paths(self) -> list[str]: ...
    def model_dump_json(self) -> str: ...
    def with_modules(self, modules: list[ModuleConfig]) -> ProjectConfig: ...
    def set_modules(self, module_paths: list[str]) -> None: ...
    def dependencies_for_module(self, module: str) -> list[DependencyConfig]: ...
    def add_dependency_to_module(
        self, module: str, dependency: DependencyConfig
    ) -> None: ...
    def compare_dependencies(
        self, other_config: ProjectConfig
    ) -> list[UnusedDependencies]: ...
    def dump_project_config_to_toml(self, project_config: ProjectConfig) -> str: ...

class TachPytestPluginHandler:
    removed_test_paths: set[str]
    all_affected_modules: set[str]
    num_removed_items: int
    tests_ran_to_completion: bool
    def __new__(
        cls,
        project_root: Path,
        project_config: ProjectConfig,
        changed_files: list[Path],
        all_affected_modules: set[Path],
    ) -> TachPytestPluginHandler: ...
    def remove_test_path(self, path: Path) -> None: ...
    def should_remove_items(self, file_path: Path) -> bool: ...



from __future__ import annotations

from copy import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from tach.errors import TachSetupError
from tach.extension import ProjectConfig, TachPytestPluginHandler
from tach.filesystem.git_ops import get_changed_files

if TYPE_CHECKING:
    from pathlib import Path


@dataclass
class AffectedTestsResult:
    exit_code: int
    tests_ran_to_completion: bool


def run_affected_tests(
    project_root: Path,
    project_config: ProjectConfig,
    head: str,
    base: str,
    pytest_args: list[Any] | None = None,
) -> AffectedTestsResult:
    try:
        import pytest  # type: ignore  # noqa: F401
    except ImportError:
        raise TachSetupError("Cannot run tests, could not find 'pytest'.")

    class TachPytestPlugin:
        def __init__(
            self,
            handler: TachPytestPluginHandler,
        ):
            self.handler = handler

        def pytest_collection_modifyitems(
            self,
            session: pytest.Session,
            config: pytest.Config,
            items: list[pytest.Item],
        ):
            seen: set[Path] = set()
            for item in copy(items):
                if not item.path:
                    continue
                if str(item.path) in self.handler.removed_test_paths:
                    self.handler.num_removed_items += 1
                    items.remove(item)
                    continue
                if item.path in seen:
                    continue

                if str(item.path) in self.handler.all_affected_modules:
                    # If this test file was changed,
                    # then we know we need to rerun it
                    seen.add(item.path)
                    continue

                if self.handler.should_remove_items(file_path=item.path.resolve()):
                    self.handler.num_removed_items += 1
                    items.remove(item)
                    self.handler.remove_test_path(item.path)

                seen.add(item.path)

        def pytest_report_collectionfinish(
            self,
            config: pytest.Config,
            start_path: Path,
            startdir: Any,
            items: list[pytest.Item],
        ) -> str | list[str]:
            return [
                f"[Tach] Skipped {len(self.handler.removed_test_paths)} test file{'s' if len(self.handler.removed_test_paths) > 1 else ''}"
                f" ({self.handler.num_removed_items} tests)"
                " since they were unaffected by current changes.",
                *(
                    f"[Tach] > Skipped '{test_path}'"
                    for test_path in self.handler.removed_test_paths
                ),
            ]

        def pytest_terminal_summary(
            self, terminalreporter: Any, exitstatus: int, config: pytest.Config
        ):
            self.handler.tests_ran_to_completion = True

    changed_files = get_changed_files(project_root, head=head, base=base)
    pytest_plugin_handler = TachPytestPluginHandler(
        project_root=project_root,
        project_config=project_config,
        changed_files=changed_files,
        all_affected_modules={changed_file.resolve() for changed_file in changed_files},
    )

    exit_code = pytest.main(
        pytest_args, plugins=[TachPytestPlugin(handler=pytest_plugin_handler)]
    )

    if exit_code == pytest.ExitCode.NO_TESTS_COLLECTED:
        # Selective testing means running zero tests will happen regularly,
        # so we do not want the default behavior of failing when no tests
        # are collected.
        exit_code = pytest.ExitCode.OK

    return AffectedTestsResult(
        exit_code=exit_code,
        tests_ran_to_completion=pytest_plugin_handler.tests_ran_to_completion,
    )


__all__ = ["run_affected_tests"]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from tach import errors
from tach.colors import BCOLORS
from tach.extension import (
    create_dependency_report,
    get_external_imports,
    set_excluded_paths,
)
from tach.filesystem import walk_pyfiles
from tach.utils.display import create_clickable_link
from tach.utils.exclude import is_path_excluded
from tach.utils.external import (
    get_package_name,
    is_stdlib_module,
    normalize_package_name,
)

if TYPE_CHECKING:
    from tach.extension import ProjectConfig


def report(
    project_root: Path,
    path: Path,
    project_config: ProjectConfig,
    include_dependency_modules: list[str] | None = None,
    include_usage_modules: list[str] | None = None,
    skip_dependencies: bool = False,
    skip_usages: bool = False,
    exclude_paths: list[str] | None = None,
) -> str:
    if not project_root.is_dir():
        raise errors.TachSetupError(
            f"The path '{project_root}' is not a valid directory."
        )

    if not path.exists():
        raise errors.TachError(f"The path '{path}' does not exist.")

    # This informs the Rust extension ahead-of-time which paths are excluded.
    set_excluded_paths(
        project_root=str(project_root),
        exclude_paths=exclude_paths or [],
        use_regex_matching=project_config.use_regex_matching,
    )

    try:
        return create_dependency_report(
            project_root=str(project_root),
            source_roots=[
                str(project_root / source_root)
                for source_root in project_config.source_roots
            ],
            path=str(path),
            include_dependency_modules=include_dependency_modules,
            include_usage_modules=include_usage_modules,
            skip_dependencies=skip_dependencies,
            skip_usages=skip_usages,
            ignore_type_checking_imports=project_config.ignore_type_checking_imports,
        )
    except ValueError as e:
        raise errors.TachError(str(e))


@dataclass
class ExternalDependency:
    absolute_file_path: Path
    import_module_path: str
    import_line_number: int
    package_name: str


def render_external_dependency(
    dependency: ExternalDependency, display_path: Path
) -> str:
    clickable_link = create_clickable_link(
        file_path=dependency.absolute_file_path,
        display_path=display_path,
        line=dependency.import_line_number,
    )
    return (
        f"{BCOLORS.OKGREEN}{clickable_link}{BCOLORS.ENDC}{BCOLORS.OKCYAN}: "
        f"Import '{dependency.import_module_path}' from package '{dependency.package_name}'{BCOLORS.ENDC}"
    )


def render_external_dependency_report(
    path: Path, dependencies: list[ExternalDependency], raw: bool = False
) -> str:
    if not dependencies:
        if raw:
            return ""
        return f"{BCOLORS.OKCYAN}No external dependencies found in {BCOLORS.ENDC}{BCOLORS.OKGREEN}'{path}'.{BCOLORS.ENDC}"

    if raw:
        return "\n".join({dependency.package_name for dependency in dependencies})

    title = f"[ External Dependencies in '{path}' ]"
    divider = "-" * len(title)
    lines = [title, divider]

    if not dependencies:
        lines.append(f"{BCOLORS.OKGREEN}No external dependencies found.{BCOLORS.ENDC}")
        return "\n".join(lines)

    for dependency in dependencies:
        lines.append(
            render_external_dependency(
                dependency=dependency,
                display_path=dependency.absolute_file_path.relative_to(Path.cwd()),
            )
        )

    return "\n".join(lines)


def get_external_dependencies(
    source_roots: list[str],
    file_path: str,
    ignore_type_checking_imports: bool,
    excluded_modules: set[str] | None = None,
) -> list[ExternalDependency]:
    external_imports = get_external_imports(
        source_roots=source_roots,
        file_path=file_path,
        ignore_type_checking_imports=ignore_type_checking_imports,
    )

    excluded_modules = excluded_modules or set()
    external_dependencies: list[ExternalDependency] = []
    for external_import in external_imports:
        external_package = get_package_name(external_import[0])
        if external_package in excluded_modules:
            continue

        if is_stdlib_module(external_package):
            continue

        external_dependencies.append(
            ExternalDependency(
                absolute_file_path=Path(file_path),
                import_module_path=external_import[0],
                import_line_number=external_import[1],
                package_name=normalize_package_name(external_import[0]),
            )
        )
    return external_dependencies


def external_dependency_report(
    project_root: Path,
    path: Path,
    project_config: ProjectConfig,
    raw: bool = False,
    exclude_paths: list[str] | None = None,
) -> str:
    if not project_root.is_dir():
        raise errors.TachSetupError(
            f"The path '{project_root}' is not a valid directory."
        )

    if not path.exists():
        raise errors.TachError(f"The path '{path}' does not exist.")

    if exclude_paths and is_path_excluded(
        exclude_paths,
        path,
        use_regex_matching=project_config.use_regex_matching,
    ):
        raise errors.TachError(f"The path '{path}' is excluded.")

    # This informs the Rust extension ahead-of-time which paths are excluded.
    set_excluded_paths(
        project_root=str(project_root),
        exclude_paths=exclude_paths or [],
        use_regex_matching=project_config.use_regex_matching,
    )
    source_roots = [
        str(project_root / source_root) for source_root in project_config.source_roots
    ]

    if path.is_file():
        external_dependencies = get_external_dependencies(
            source_roots=source_roots,
            file_path=str(path.resolve()),
            excluded_modules=set(project_config.external.exclude),
            ignore_type_checking_imports=project_config.ignore_type_checking_imports,
        )
        return render_external_dependency_report(path, external_dependencies, raw=raw)

    all_external_dependencies: list[ExternalDependency] = []
    for pyfile in walk_pyfiles(
        path,
        project_root=project_root,
        exclude_paths=exclude_paths,
        use_regex_matching=project_config.use_regex_matching,
    ):
        all_external_dependencies.extend(
            get_external_dependencies(
                source_roots=source_roots,
                file_path=str(path.resolve() / pyfile),
                excluded_modules=set(project_config.external.exclude),
                ignore_type_checking_imports=project_config.ignore_type_checking_imports,
            )
        )

    return render_external_dependency_report(path, all_external_dependencies, raw=raw)


__all__ = ["report", "external_dependency_report"]



from __future__ import annotations

from tach.filesystem.install import install_pre_commit
from tach.filesystem.project import (
    build_project_config_path,
    find_project_config_root,
    get_deprecated_project_config_path,
    get_project_config_path,
)
from tach.filesystem.service import (
    file_to_module_path,
    module_to_pyfile_or_dir_path,
    walk,
    walk_pyfiles,
    write_file,
)

__all__ = [
    "write_file",
    "walk",
    "walk_pyfiles",
    "file_to_module_path",
    "module_to_pyfile_or_dir_path",
    "get_project_config_path",
    "build_project_config_path",
    "get_deprecated_project_config_path",
    "find_project_config_root",
    "install_pre_commit",
]



from __future__ import annotations

import json
from copy import copy
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING
from urllib import error, request

from tach import filesystem as fs
from tach.extension import ModuleConfig, ProjectConfig

if TYPE_CHECKING:
    from pathlib import Path

    import pydot  # type: ignore


TACH_SHOW_URL = "https://show.gauge.sh"


def generate_show_url(
    project_root: Path,
    project_config: ProjectConfig,
    included_paths: list[Path] | None = None,
) -> str | None:
    if included_paths:
        project_config = filter_project_config(
            project_config, project_root=project_root, included_paths=included_paths
        )
    json_data = project_config.model_dump_json()
    json_bytes = json_data.encode("utf-8")
    req = request.Request(
        f"{TACH_SHOW_URL}/api/core/0.6.5/graph/",
        data=json_bytes,
        headers={"Content-Type": "application/json"},
    )

    try:
        # Send the request and read the response
        with request.urlopen(req) as response:
            response_data = response.read().decode("utf-8")
            response_json = json.loads(response_data)
            url = response_json.get("uid")
            return f"{TACH_SHOW_URL}?uid={url}"
    except (UnicodeDecodeError, JSONDecodeError, error.URLError) as e:
        print(f"Error: {e}")
        return None


def module_path_is_included_in_paths(
    source_roots: tuple[Path, ...], module_path: str, included_paths: list[Path] | None
) -> bool:
    if included_paths is None:
        return True
    module_fs_path = fs.module_to_pyfile_or_dir_path(source_roots, module_path)
    if not module_fs_path:
        return False
    for included_path in included_paths:
        if included_path in module_fs_path.parents or included_path == module_fs_path:
            return True
    return False


def filter_project_config(
    project_config: ProjectConfig,
    project_root: Path,
    included_paths: list[Path],
) -> ProjectConfig:
    source_roots = tuple(
        map(lambda source_root: project_root / source_root, project_config.source_roots)
    )
    included_paths = list(map(lambda path: project_root / path, included_paths))
    all_modules = copy(project_config.modules)
    project_config.set_modules([])
    filtered_modules: list[ModuleConfig] = []
    for module in all_modules:
        if module_path_is_included_in_paths(source_roots, module.path, included_paths):
            filtered_modules.append(module)
    return ProjectConfig.with_modules(project_config, filtered_modules)


def generate_module_graph_dot_file(
    project_root: Path,
    project_config: ProjectConfig,
    output_filepath: Path,
    included_paths: list[Path] | None = None,
) -> None:
    # Local import because networkx takes about ~100ms to load
    import networkx as nx

    graph = nx.DiGraph()  # type: ignore

    def upsert_edge(graph: nx.DiGraph, module: str, dependency: str) -> None:  # type: ignore
        if module not in graph:
            graph.add_node(module)  # type: ignore
        if dependency not in graph:
            graph.add_node(dependency)  # type: ignore
        graph.add_edge(module, dependency)  # type: ignore

    source_roots = tuple(
        map(lambda source_root: project_root / source_root, project_config.source_roots)
    )
    included_paths = (
        list(map(lambda path: project_root / path, included_paths))
        if included_paths
        else None
    )
    for module in project_config.modules:
        module_is_included = module_path_is_included_in_paths(
            source_roots, module.path, included_paths
        )
        for dependency in module.depends_on:
            dependency_is_included = module_path_is_included_in_paths(
                source_roots, dependency.path, included_paths
            )

            # This essentially means we propagate one degree from the included modules
            if module_is_included or dependency_is_included:
                upsert_edge(graph, module.path, dependency.path)  # type: ignore

    pydot_graph: pydot.Dot = nx.nx_pydot.to_pydot(graph)  # type: ignore
    dot_data: str = pydot_graph.to_string()  # type: ignore

    output_filepath.write_text(dot_data)  # type: ignore


def generate_module_graph_mermaid(
    project_root: Path,
    project_config: ProjectConfig,
    output_filepath: Path,
    included_paths: list[Path] | None = None,
):
    if included_paths:
        project_config = filter_project_config(
            project_config, project_root, included_paths
        )
    edges: list[str] = []
    isolated: list[str] = []
    for module in project_config.modules:
        for dependency in module.depends_on:
            edges.append(
                f"    {module.path.strip('<>')} --> {dependency.path.strip('<>')}"
            )
        if not module.depends_on:
            isolated.append(f"    {module.path.strip('<>')}")

    mermaid_graph = "graph TD\n" + "\n".join(edges) + "\n" + "\n".join(isolated)

    output_filepath.write_text(mermaid_graph)


__all__ = [
    "generate_show_url",
    "generate_module_graph_dot_file",
    "generate_module_graph_mermaid",
]



from __future__ import annotations

PACKAGE_NAME = "tach"
TOOL_NAME = "tach"
CONFIG_FILE_NAME = TOOL_NAME
PACKAGE_FILE_NAME = "package"
ROOT_MODULE_SENTINEL_TAG = "<root>"
TACH_YML_SCHEMA_URL = "https://raw.githubusercontent.com/gauge-sh/tach/v0.12.0/public/tach-yml-schema.json"

DEFAULT_EXCLUDE_PATHS = ["tests", "docs", ".*__pycache__", ".*egg-info"]

__all__ = [
    "PACKAGE_NAME",
    "TOOL_NAME",
    "CONFIG_FILE_NAME",
    "PACKAGE_FILE_NAME",
    "ROOT_MODULE_SENTINEL_TAG",
    "TACH_YML_SCHEMA_URL",
    "DEFAULT_EXCLUDE_PATHS",
]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import tomli
import tomli_w

from tach import filesystem as fs
from tach.extension import (
    dump_project_config_to_toml as ext_dump_project_config_to_toml,
)
from tach.extension import (
    parse_project_config as ext_parse_project_config,
)

if TYPE_CHECKING:
    from tach.extension import ProjectConfig


def dump_project_config_to_toml(config: ProjectConfig) -> str:
    data = tomli.loads(ext_dump_project_config_to_toml(config))
    return tomli_w.dumps(data)


def migrate_deprecated_config(filepath: Path) -> ProjectConfig:
    import yaml

    content = filepath.read_text()
    data = yaml.safe_load(content)

    try:
        if "cache" in data:
            if "backend" in data["cache"]:
                # Force cache backend to 'disk' (original value was 'local')
                data["cache"]["backend"] = "disk"
        # Old migrations
        if "modules" in data:
            for module in data["modules"]:
                if "depends_on" in module:
                    for index, path in enumerate(module["depends_on"]):
                        if isinstance(path, str):
                            module["depends_on"][index] = {"path": path}
        if "source_root" in data and isinstance(data["source_root"], str):
            data["source_roots"] = [data["source_root"]]
            del data["source_root"]
        toml_config = tomli_w.dumps(data)
        print("Auto-migrating deprecated YAML config to TOML...")
        filepath.with_suffix(".toml").write_text(toml_config)
        project_config = ext_parse_project_config(filepath.with_suffix(".toml"))
    except TypeError as e:
        raise ValueError(f"Failed to parse deprecated YAML config: {e}")
    except ValueError as e:
        filepath.with_suffix(".toml").unlink()
        raise ValueError(f"Failed to parse deprecated YAML config: {e}")
    print("Deleting deprecated YAML config...")
    filepath.unlink()
    return project_config


def parse_project_config(root: Path | None = None) -> ProjectConfig | None:
    root = root or Path.cwd()
    file_path = fs.get_project_config_path(root)
    if file_path:
        # Standard TOML config found
        project_config = ext_parse_project_config(file_path)
    else:
        # No TOML found, check for deprecated (YAML) config as a fallback
        file_path = fs.get_deprecated_project_config_path(root)
        if not file_path:
            return None
        # Return right away, this is a final ProjectConfig
        project_config = migrate_deprecated_config(file_path)
    return project_config
